"""Converts an OSeMOSYS datafile to a Tabular Data Package

"""
import logging
import os
import sys
from typing import Dict, List

import pandas as pd
from amply import Amply
from flatten_dict import flatten

from otoole import read_packaged_file
from otoole.preprocess.longify_data import (
    check_datatypes,
    check_set_datatype,
    write_out_dataframe,
)

logger = logging.getLogger(__name__)


def write_default_values(filepath):

    config = read_packaged_file("config.yaml", "otoole.preprocess")

    default_values_path = os.path.join(filepath, "data", "default_values.csv")
    with open(default_values_path, "w") as csv_file:
        csv_file.write("name,default_value\n")

        for name, contents in config.items():
            if contents["type"] == "param":
                csv_file.write("{},{}\n".format(name, contents["default"]))


def convert_file_to_package(path_to_datafile: str, path_to_datapackage: str):
    """Converts an OSeMOSYS datafile to a Tabular Data Package

    Arguments
    ---------
    path_to_datafile: str
    path_to_datapackage: str
        Path to the folder in which to write the new Tabular Data Package

    """
    dict_of_dataframes = read_datafile_to_dict(path_to_datafile)
    if not os.path.exists(path_to_datapackage):
        os.mkdir(path_to_datapackage)
    for name, df in dict_of_dataframes.items():
        write_out_dataframe(path_to_datapackage, name, df)
    datapackage = read_packaged_file("datapackage.json", "otoole.preprocess")
    filepath = os.path.join(path_to_datapackage, "datapackage.json")
    with open(filepath, "w") as destination:
        destination.writelines(datapackage)
    write_default_values(path_to_datapackage)


def read_datafile_to_dict(path_to_datafile: str) -> Dict[str, pd.DataFrame]:
    """Reads in an OSeMOSYS GNUMathProg datafile into *otoole* format of choice

    Arguments
    ---------
    path_to_datafile: str
    """
    config = read_packaged_file("config.yaml", "otoole.preprocess")  # type: Dict
    amply_datafile = read_in_datafile(path_to_datafile, config)
    return convert_amply_to_dataframe(amply_datafile, config)


def read_in_datafile(path_to_datafile: str, config: Dict) -> Amply:
    """Read in a datafile using the Amply parsing class

    Arguments
    ---------
    path_to_datafile: str
    config: Dict
    """
    parameter_definitions = load_parameter_definitions(config)
    datafile_parser = Amply(parameter_definitions)
    logger.debug(datafile_parser)

    with open(path_to_datafile, "r") as datafile:
        datafile_parser.load_file(datafile)

    return datafile_parser


def convert_amply_to_dataframe(
    datafile_parser: Amply, config: Dict
) -> Dict[str, pd.DataFrame]:
    """Converts an amply parser to a dict of pandas dataframes

    Arguments
    ---------
    datafile_parser : Amply
    config : Dict

    Returns
    -------
    dict of pandas.DataFrame
    """

    dict_of_dataframes = {}

    for name in datafile_parser.symbols.keys():
        logger.debug("Extracting data for %s", name)
        if config[name]["type"] == "param":
            indices = config[name]["indices"]
            indices_dtypes = [config[index]["dtype"] for index in indices]
            indices.append("VALUE")
            indices_dtypes.append("float")

            raw_data = datafile_parser[name].data
            data = convert_amply_data_to_list(raw_data)
            df = pd.DataFrame(data=data, columns=indices)
            try:
                dict_of_dataframes[name] = check_datatypes(df, config, name)
            except ValueError as ex:
                msg = "Validation error when checking datatype of {}: {}".format(
                    name, str(ex)
                )
                raise ValueError(msg)
        elif config[name]["type"] == "set":
            data = datafile_parser[name].data
            logger.debug(data)

            indices = ["VALUE"]
            df = pd.DataFrame(data=data, columns=indices, dtype=config[name]["dtype"])
            dict_of_dataframes[name] = check_set_datatype(df, config, name)
        logger.debug("\n%s\n", dict_of_dataframes[name])

    return dict_of_dataframes


def convert_amply_data_to_list(amply_data: Dict) -> List[List]:
    """Flattens a dictionary into a list of lists

    Arguments
    ---------
    amply_data: dict
    """

    data = []

    raw_data = flatten(amply_data)
    for key, value in raw_data.items():
        data.append(list(key) + [value])

    return data


def load_parameter_definitions(config: dict) -> str:
    """Load the set and parameter dimensions into datafile parser

    Returns
    -------
    str
    """
    elements = ""

    for name, attributes in config.items():
        if attributes["type"] == "param":
            elements += "param {} {};\n".format(
                name, "{" + ",".join(attributes["indices"]) + "}"
            )
        elif attributes["type"] == "set":
            elements += "set {};\n".format(name)

    logger.debug(elements)
    return elements


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    path_to_datafile = sys.argv[1]
    path_to_datapackage = sys.argv[2]

    convert_file_to_package(path_to_datafile, path_to_datapackage)



import io
from unittest.mock import Mock

from pytest import fixture

import pandas as pd

from otoole.preprocess.excel_to_osemosys import read_config
from otoole.preprocess.narrow_to_datafile import (
    DataPackageToDatafile,
    DataPackageToExcel,
)


class TestDataFrameWritingExcel:
    @fixture
    def setup(self, monkeypatch) -> DataPackageToExcel:

        dp = DataPackageToExcel
        monkeypatch.setattr(dp, "_get_package", Mock())  # type: ignore
        monkeypatch.setattr(dp, "_get_default_values", Mock())  # type: ignore

        return dp("", "")

    def test_form_empty_parameter_with_defaults(self, setup):

        convert = setup  # typing: DataPackageToExcel

        data = []

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])
        actual = convert._form_parameter(df, "test_parameter", 0)
        expected = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_one_columns(self, setup):

        convert = setup  # typing: DataPackageToExcel

        data = ["A", "B", "C"]

        df = pd.DataFrame(data=data, columns=["FUEL"])
        actual = convert._form_parameter(df, "test_set", 0)
        expected = pd.DataFrame(data=["A", "B", "C"], columns=["FUEL"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_three_columns(self, setup):

        convert = setup  # typing: DataPackageToExcel

        data = [["SIMPLICITY", "COAL", 2015, 41], ["SIMPLICITY", "COAL", 2016, 42]]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        actual = convert._form_parameter(df, "test_parameter", 0)

        expected_data = [[41, 42]]
        expected = pd.DataFrame(
            data=expected_data,
            columns=pd.Int64Index([2015, 2016], dtype="int64", name="YEAR"),
            index=pd.MultiIndex.from_tuples(
                [("SIMPLICITY", "COAL")], names=["REGION", "FUEL"]
            ),
        )

        pd.testing.assert_frame_equal(actual, expected)


class TestDataFrameWritingDatafile:
    @fixture
    def setup(self, monkeypatch) -> DataPackageToDatafile:

        dp = DataPackageToDatafile
        monkeypatch.setattr(dp, "_get_package", Mock())  # type: ignore
        monkeypatch.setattr(dp, "_get_default_values", Mock())  # type: ignore
        return dp("", "")

    def test_write_empty_parameter_with_defaults(self, setup):

        convert = setup  # typing: DataPackageToCsv

        data = []

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])

        stream = io.StringIO()
        convert._write_parameter(df, "test_parameter", stream, 0)

        stream.seek(0)
        expected = ["param default 0 : test_parameter :=\n", ";\n"]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_parameter_as_tabbing_format(self, setup):

        data = [["SIMPLICITY", "BIOMASS", 0.95969], ["SIMPLICITY", "ETH1", 4.69969]]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])

        stream = io.StringIO()
        convert = setup
        convert._write_parameter(df, "test_parameter", stream, 0)

        stream.seek(0)
        expected = [
            "param default 0 : test_parameter :=\n",
            "SIMPLICITY BIOMASS 0.95969\n",
            "SIMPLICITY ETH1 4.69969\n",
            ";\n",
        ]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_parameter_skip_defaults(self, setup):

        data = [
            ["SIMPLICITY", "BIOMASS", 0.95969],
            ["SIMPLICITY", "ETH1", 4.69969],
            ["SIMPLICITY", "ETH2", -1],
            ["SIMPLICITY", "ETH3", -1],
        ]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])

        stream = io.StringIO()
        convert = setup
        convert._write_parameter(df, "test_parameter", stream, -1)

        stream.seek(0)
        expected = [
            "param default -1 : test_parameter :=\n",
            "SIMPLICITY BIOMASS 0.95969\n",
            "SIMPLICITY ETH1 4.69969\n",
            ";\n",
        ]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_set(self, setup):

        data = [["BIOMASS"], ["ETH1"]]

        df = pd.DataFrame(data=data, columns=["VALUE"])

        stream = io.StringIO()
        convert = setup
        convert._write_set(df, "TECHNOLOGY", stream)

        stream.seek(0)
        expected = ["set TECHNOLOGY :=\n", "BIOMASS\n", "ETH1\n", ";\n"]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line


class TestConfig:
    def test_read_config(self):

        actual = read_config()
        expected = {
            "default": 0,
            "dtype": "float",
            "indices": ["REGION", "FUEL", "YEAR"],
            "type": "param",
        }
        assert actual["AccumulatedAnnualDemand"] == expected



"""Read in a folder of irregular wide-format csv files and write them out as narrow csvs
"""
import logging
import os
import sys
from typing import Dict

import pandas as pd

from otoole import read_packaged_file

logger = logging.getLogger()


def check_parameter(df, config_details, name):
    actual_headers = df.columns
    expected_headers = config_details["indices"]
    logger.debug("Expected headers for %s: %s", name, expected_headers)

    if "REGION" in expected_headers and "REGION" not in actual_headers:
        logger.info("Added 'REGION' column to %s", name)
        df["REGION"] = "SIMPLICITY"

    if "MODEOFOPERATION" in actual_headers:
        df = df.rename(columns={"MODEOFOPERATION": "MODE_OF_OPERATION"})

    if actual_headers[-1] == "VALUE":
        logger.info("%s is already in narrow form with headers %s", name, df.columns)
        narrow = df
    else:
        try:
            narrow = pd.melt(
                df,
                id_vars=expected_headers[:-1],
                var_name=expected_headers[-1],
                value_name="VALUE",
            )
        except IndexError as ex:
            logger.debug(df.columns)
            raise ex

    expected_headers.append("VALUE")
    for column in expected_headers:
        if column not in narrow.columns:
            logger.warning("%s not in header of %s", column, name)

    logger.debug("Final expected headers for %s: %s", name, expected_headers)

    return narrow[expected_headers]


def check_set(df, config_details, name):

    logger.info("Checking set %s", name)
    narrow = df

    return narrow


def check_set_datatype(
    narrow: pd.DataFrame, config_details: Dict, set_name: str
) -> pd.DataFrame:
    """Checks the datatypes of a set_name dataframe

    Arguments
    ---------
    narrow : pandas.DataFrame
        The set data
    config_details : dict
        The configuration dictionary
    set_name : str
        The name of the set
    """
    datatype = config_details[set_name]["dtype"]
    logger.debug("Columns for set %s are: %s", set_name, narrow.columns)
    if narrow.iloc[:, 0].dtype != datatype:
        logger.warning("dtype does not match %s for set %s", datatype, set_name)
    return narrow


def check_datatypes(
    narrow: pd.DataFrame, config_details: Dict, parameter: str
) -> pd.DataFrame:
    """Checks a parameters datatypes

    Arguments
    ---------
    narrow : pandas.DataFrame
        The parameter data
    config_details : dict
        The configuration dictionary
    parameter : str
        The name of the parameter
    """
    logger.info("Checking datatypes for %s", parameter)
    dtypes = {}

    for column in narrow.columns:
        if column == "VALUE":
            datatype = config_details[parameter]["dtype"]
            dtypes["VALUE"] = datatype
        else:
            datatype = config_details[column]["dtype"]
            dtypes[column] = datatype
        if narrow[column].dtype != datatype:
            logger.warning(
                "dtype of column %s does not match %s for parameter %s",
                column,
                datatype,
                parameter,
            )
            if datatype == "int":
                try:
                    narrow[column] = narrow[column].apply(_cast_to_int)
                except ValueError as ex:
                    msg = "Unable to apply datatype for column {}: {}".format(
                        column, str(ex)
                    )
                    raise ValueError(msg)
    return narrow.astype(dtypes)


def _cast_to_int(value):
    return int(float(value))


def main(output_folder: str, narrow_folder: str):
    """Read in a folder of irregular wide-format files and write as narrow csvs

    Arguments
    ---------
    output_folder: str
        The source folder of wide files
    narrow_folder: str
        The destination folder

    """
    config = read_packaged_file("config.yaml", "otoole.preprocess")

    for parameter, details in config.items():
        logger.info("Looking for %s", parameter)
        config_details = config[parameter]

        filepath = os.path.join(output_folder, parameter + ".csv")

        try:
            df = pd.read_csv(filepath)
        except pd.errors.EmptyDataError:
            logger.error("No data found in file for %s", parameter)
            expected_columns = config_details["indices"]
            default_columns = expected_columns + ["VALUE"]
            df = pd.DataFrame(columns=default_columns)

        entity_type = config[parameter]["type"]

        if entity_type == "param":
            narrow = check_parameter(df, config_details, parameter)
            if not narrow.empty:
                narrow_checked = check_datatypes(narrow, config, parameter)
            else:
                narrow_checked = narrow
        elif entity_type == "set":
            narrow = check_set(df, config_details, parameter)
            if not narrow.empty:
                narrow_checked = check_set_datatype(narrow, config, parameter)
            else:
                narrow_checked = narrow

        write_out_dataframe(narrow_folder, parameter, narrow_checked)


def write_out_dataframe(folder, parameter, df):
    """Writes out a dataframe as a csv into the data subfolder of a datapackage

    Arguments
    ---------
    folder : str
    parameter : str
    df : pandas.DataFrame

    """
    os.makedirs(os.path.join(folder, "data"), exist_ok=True)
    filepath = os.path.join(folder, "data", parameter + ".csv")
    with open(filepath, "w") as csvfile:
        logger.info("Writing %s rows into narrow file for %s", df.shape[0], parameter)
        df.to_csv(csvfile, index=False)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    output_folder = sys.argv[1]
    narrow_folder = sys.argv[2]
    main(output_folder, narrow_folder)



"""Converts an OSeMOSYS solution file from CPLEX, CBC or GLPK into CBC or CSV format

"""
import argparse
import logging
import os
from typing import Dict, List, Optional, Tuple

import pandas as pd
from pandas_datapackage_reader import read_datapackage

from otoole import read_packaged_file
from otoole.preprocess import read_datafile_to_dict
from otoole.results.result_package import ResultsPackage

LOGGER = logging.getLogger(__name__)


class ConvertLine(object):
    """Abstract class which defines the interface to the family of convertors

    Inherit this class and implement the ``_do_it()`` method to produce the
    data to be written out into a new format

    Example
    -------
    >>> cplex_line = "AnnualCost	REGION	CDBACKSTOP	1.0	0.0	137958.8400384134"
    >>> convertor = RegionTechnology()
    >>> convertor.convert()
    VariableName(REGION,TECHCODE01,2015)       42.69         0\\n
    VariableName(REGION,TECHCODE01,2017)       137958.84         0\\n
    """

    def __init__(self, data: List, start_year: int, end_year: int, output_format="cbc"):
        self.data = data
        self.start_year = start_year
        self.end_year = end_year
        self.output_format = output_format

    def _do_it(self) -> Tuple:
        raise NotImplementedError()

    def convert(self) -> List[str]:
        if self.output_format == "cbc":
            convert = self.convert_cbc()
        elif self.output_format == "csv":
            convert = self.convert_csv()
        return convert

    def convert_csv(self) -> List[str]:
        """Format the data for writing to a csv file
        """
        data = []
        variable, dimensions, values = self._do_it()

        for index, value in enumerate(values):

            year = self.start_year + index
            if (value not in ["0.0", "0", ""]) and (year <= self.end_year):

                try:
                    value = float(value)
                except ValueError:
                    value = 0

                full_dims = ",".join(dimensions + (str(year),))

                formatted_data = '{0},"{1}",{2}\n'.format(variable, full_dims, value)

                data.append(formatted_data)

        return data

    def convert_cbc(self) -> List[str]:
        """Format the data for writing to a CBC file
        """
        cbc_data = []
        variable, dimensions, values = self._do_it()

        for index, value in enumerate(values):

            year = self.start_year + index
            if (value not in ["0.0", "0", ""]) and (year <= self.end_year):

                try:
                    value = float(value)
                except ValueError:
                    value = 0

                full_dims = ",".join(dimensions + (str(year),))

                formatted_data = "0 {0}({1}) {2} 0\n".format(variable, full_dims, value)

                cbc_data.append(formatted_data)

        return cbc_data


class RegionTimeSliceTechnologyMode(ConvertLine):
    def _do_it(self) -> Tuple:
        """Produces output indexed by Region, Timeslice, Tech and Mode

        ``0 VariableName(REGION,SD1D,TECHCODE01,2,2015) 42.69 0\n``

        """
        variable = self.data[0]
        region = self.data[1]
        timeslice = self.data[2]
        technology = self.data[3]
        mode = self.data[4]
        values = self.data[5:]

        dimensions = (region, timeslice, technology, mode)

        return (variable, dimensions, values)


class RegionTechnology(ConvertLine):
    def _do_it(self) -> Tuple:
        """Produces output indexed by dimensions Region and Technology

        ``0 VariableName(REGION,TECHCODE01,2015) 42.69 0\n``

        """
        variable = self.data[0]
        region = self.data[1]
        technology = self.data[2]

        dimensions = (region, technology)

        values = self.data[3:]

        return (variable, dimensions, values)


def process_line(
    line: str, start_year: int, end_year: int, output_format: str
) -> List[str]:
    """Processes an individual line in a CPLEX file

    A different ConvertLine implementation is chosen depending upon the
    variable name

    Arguments
    ---------
    line: str
    start_year: int
    end_year: int
    output_format: str
        The file format required - either ``csv`` or ``cbc``
    """
    row_as_list = line.split("\t")
    variable = row_as_list[0]
    if variable in [
        "NewCapacity",
        "TotalCapacityAnnual",
        "CapitalInvestment",
        "AnnualFixedOperatingCost",
        "AnnualVariableOperatingCost",
    ]:
        convertor = RegionTechnology(
            row_as_list, start_year, end_year, output_format
        ).convert()
    elif variable in ["RateOfActivity"]:
        convertor = RegionTimeSliceTechnologyMode(
            row_as_list, start_year, end_year, output_format
        ).convert()
    else:
        convertor = []

    return convertor


def convert_cplex_file(
    cplex_filename: str,
    output_filename: str,
    start_year=2015,
    end_year=2070,
    output_format="cbc",
):
    """Converts a CPLEX solution file into that of the CBC solution file

    Arguments
    ---------
    cplex_filename : str
        Path to the transformed CPLEX solution file
    output_filename : str
        Path for the processed data to be written to
    """

    with open(output_filename, "w") as cbc_file:
        with open(cplex_filename, "r") as cplex_file:
            for linenum, line in enumerate(cplex_file):
                try:
                    convertor = process_line(line, start_year, end_year, output_format)
                    if convertor:
                        cbc_file.writelines(convertor)
                except ValueError:
                    msg = "Error caused at line {}: {}"
                    raise ValueError(msg.format(linenum, line))


def convert_cbc_to_dataframe(data_file: str) -> pd.DataFrame:
    """Reads a CBC solution file into a pandas DataFrame

    Arguments
    ---------
    data_file : str
    """
    df = pd.read_csv(
        data_file,
        header=None,
        names=["temp", "VALUE"],
        delim_whitespace=True,
        skiprows=1,
        usecols=[1, 2],
    )  # type: pd.DataFrame
    df.columns = ["temp", "Value"]
    df[["Variable", "Index"]] = df["temp"].str.split("(", expand=True)
    df = df.drop("temp", axis=1)
    df["Index"] = df["Index"].str.replace(")", "")
    return df[["Variable", "Index", "Value"]]


def convert_dataframe_to_csv(
    data: pd.DataFrame, input_data: Optional[Dict[str, pd.DataFrame]] = None
) -> Dict[str, pd.DataFrame]:
    """Convert from dataframe to csv

    Converts a pandas DataFrame containing all CBC results to reformatted
    dictionary of pandas DataFrames in long format ready to write out as
    csv files

    Arguments
    ---------
    data : pandas.DataFrame
        CBC results stored in a dataframe
    input_data_path : str, default=None
        Path to the OSeMOSYS data file containing input data

    Example
    -------
    >>> df = pd.DataFrame(data=[
            ['TotalDiscountedCost', "SIMPLICITY,2015", 187.01576],
            ['TotalDiscountedCost', "SIMPLICITY,2016", 183.30788]],
            columns=['Variable', 'Index', 'Value'])
    >>> convert_dataframe_to_csv(df)
    {'TotalDiscountedCost':        REGION  YEAR      VALUE
                                0  SIMPLICITY  2015  187.01576
                                1  SIMPLICITY  2016  183.30788}
    """
    input_config = read_packaged_file("config.yaml", "otoole.preprocess")
    results_config = read_packaged_file("config.yaml", "otoole.results")

    sets = {x: y for x, y in input_config.items() if y["type"] == "set"}

    results = {}  # type: Dict[str, pd.DataFrame]

    not_found = []

    for name, details in results_config.items():
        df = data[data["Variable"] == name]

        if not df.empty:

            LOGGER.debug("Extracting results for %s", name)
            indices = details["indices"]

            df[indices] = df["Index"].str.split(",", expand=True)

            types = {index: sets[index]["dtype"] for index in indices}
            df = df.astype(types)

            df = df.drop(columns=["Variable", "Index"])

            df = df.rename(columns={"Value": "VALUE"})

            results[name] = df[indices + ["VALUE"]].set_index(indices)

        else:
            not_found.append(name)

    LOGGER.debug("Unable to find CBC variables for: %s", ", ".join(not_found))

    results_package = ResultsPackage(results, input_data)

    for name in not_found:

        LOGGER.info("Looking for %s", name)
        details = results_config[name]

        try:
            df = results_package[name]
        except KeyError as ex:
            LOGGER.info("No calculation method available for %s", name)
            LOGGER.debug("Error calculating %s: %s", name, str(ex))
            df = pd.DataFrame()

        if not df.empty:
            results[name] = df
        else:
            LOGGER.warning(
                "Calculation returned empty dataframe for parameter '%s'", name
            )

    return results


def write_csvs(results_path: str, results: Dict[str, pd.DataFrame]):
    """Write out CSV files from CBC file

    Arguments
    ---------
    results_path : str
    results : dict
    """
    for name, df in results.items():
        filename = os.path.join(results_path, name + ".csv")

        if not os.path.exists(results_path):
            LOGGER.info("Creating new results folder at '%s'", results_path)
            os.makedirs(results_path, exist_ok=True)

        if not df.empty:
            df.to_csv(filename, index=True)
        else:
            LOGGER.warning("Result parameter %s is empty", name)


def convert_cbc_to_csv(
    from_file: str,
    to_file: str,
    input_data_path: str = None,
    input_data_format="datapackage",
):
    """

    Arguments
    ---------
    from_file: str
        CBC solution file
    to_file: str
        Path to directory in which CSV files will be written
    input_data_path: str
        Optional path to input data (required if using short or fast versions
        of OSeMOSYS)
    input_data_format : str, default='datapackage

    """
    if input_data_format == "datapackage" and input_data_path:
        input_data = read_datapackage(input_data_path)
    elif input_data_format == "datafile" and input_data_path:
        input_data = read_datafile_to_dict(input_data_path)
    else:
        input_data = None
    df = convert_cbc_to_dataframe(from_file)
    csv = convert_dataframe_to_csv(df, input_data)
    write_csvs(to_file, csv)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert OSeMOSYS CPLEX files into different formats"
    )
    parser.add_argument(
        "cplex_file", help="The filepath of the OSeMOSYS cplex output file"
    )
    parser.add_argument(
        "output_file", help="The filepath of the converted file that will be written"
    )
    parser.add_argument(
        "-s",
        "--start_year",
        type=int,
        default=2015,
        help="Output only the results from this year onwards",
    )
    parser.add_argument(
        "-e",
        "--end_year",
        type=int,
        default=2070,
        help="Output only the results upto and including this year",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--csv",
        action="store_true",
        help="Output file in comma-separated-values format",
    )
    group.add_argument(
        "--cbc", action="store_true", help="Output file in CBC format, (default option)"
    )

    args = parser.parse_args()

    if args.csv:
        output_format = "csv"
    else:
        output_format = "cbc"

    convert_cplex_file(
        args.cplex_file, args.output_file, args.start_year, args.end_year, output_format
    )



from .excel_to_osemosys import generate_csv_from_excel
from .create_datapackage import main as create_datapackage, csv_to_datapackage
from .narrow_to_datafile import (
    convert_datapackage_to_datafile,
    convert_datapackage_to_excel,
)
from .datafile_to_datapackage import convert_file_to_package, read_datafile_to_dict

__all__ = [
    "generate_csv_from_excel",
    "create_datapackage",
    "convert_datapackage_to_datafile",
    "convert_file_to_package",
    "create_datapackage_from_datafile",
    "convert_datapackage_to_excel",
    "read_datafile_to_dict",
    "csv_to_datapackage",
]



import tempfile

from otoole.visualise import create_res

SIMPLICITY_VERSION = 'v0.1a0'


def test_create_res():

    url = 'https://github.com/OSeMOSYS/simplicity/archive/{}.zip'.format(SIMPLICITY_VERSION)
    _, path_to_resfile = tempfile.mkstemp(suffix='.pdf')
    create_res(url, path_to_resfile)



from io import StringIO
from textwrap import dedent
from typing import List

from pytest import mark

import pandas as pd

from otoole.results.convert import (
    convert_cbc_to_dataframe,
    convert_dataframe_to_csv,
    process_line,
)


class TestCbcSoltoDataFrame:

    test_data = [
        (
            dedent(
                """Optimal - objective value 4483.96932237
                             1 TotalDiscountedCost(SIMPLICITY,2015)                                                   187.01576                       0
                             2 TotalDiscountedCost(SIMPLICITY,2016)                                                   183.30788                       0
                             3 TotalDiscountedCost(SIMPLICITY,2017)                                                   181.05465                       0
                             4 TotalDiscountedCost(SIMPLICITY,2018)                                                   218.08923                       0
                             5 TotalDiscountedCost(SIMPLICITY,2019)                                                   193.85792                       0
                             6 TotalDiscountedCost(SIMPLICITY,2020)                                                   233.79202                       0

                         """
            ),
            pd.DataFrame(
                data=[
                    ["TotalDiscountedCost", "SIMPLICITY,2015", 187.01576],
                    ["TotalDiscountedCost", "SIMPLICITY,2016", 183.30788],
                    ["TotalDiscountedCost", "SIMPLICITY,2017", 181.05465],
                    ["TotalDiscountedCost", "SIMPLICITY,2018", 218.08923],
                    ["TotalDiscountedCost", "SIMPLICITY,2019", 193.85792],
                    ["TotalDiscountedCost", "SIMPLICITY,2020", 233.79202],
                ],
                columns=["Variable", "Index", "Value"],
            ),
        )
    ]

    @mark.parametrize("cbc_input,expected", test_data, ids=["TotalDiscountedCost"])
    def test_read_cbc_to_dataframe(self, cbc_input, expected):
        with StringIO(cbc_input) as file_buffer:
            actual = convert_cbc_to_dataframe(file_buffer)
            pd.testing.assert_frame_equal(actual, expected)


class TestCbctoCsv:
    test_data = [
        # First case
        (
            pd.DataFrame(
                data=[
                    ["TotalDiscountedCost", "SIMPLICITY,2015", 187.01576],
                    ["TotalDiscountedCost", "SIMPLICITY,2016", 183.30788],
                    ["TotalDiscountedCost", "SIMPLICITY,2017", 181.05465],
                    ["TotalDiscountedCost", "SIMPLICITY,2018", 218.08923],
                    ["TotalDiscountedCost", "SIMPLICITY,2019", 193.85792],
                    ["TotalDiscountedCost", "SIMPLICITY,2020", 233.79202],
                ],
                columns=["Variable", "Index", "Value"],
            ),
            {},
            (
                {
                    "TotalDiscountedCost": pd.DataFrame(
                        data=[
                            ["SIMPLICITY", 2015, 187.01576],
                            ["SIMPLICITY", 2016, 183.30788],
                            ["SIMPLICITY", 2017, 181.05465],
                            ["SIMPLICITY", 2018, 218.08923],
                            ["SIMPLICITY", 2019, 193.85792],
                            ["SIMPLICITY", 2020, 233.79202],
                        ],
                        columns=["REGION", "YEAR", "VALUE"],
                    ).set_index(["REGION", "YEAR"])
                }
            ),
        ),
        # Second case
        (
            pd.DataFrame(
                data=[
                    ["AnnualEmissions", "REGION,CO2,2017", 137958.8400384134],
                    ["AnnualEmissions", "REGION,CO2,2018", 305945.3841061913],
                    ["AnnualEmissions", "REGION,CO2,2019", 626159.9611543404],
                ],
                columns=["Variable", "Index", "Value"],
            ),
            {},
            {
                "AnnualEmissions": pd.DataFrame(
                    data=[
                        ["REGION", "CO2", 2017, 137958.8400384134],
                        ["REGION", "CO2", 2018, 305945.3841061913],
                        ["REGION", "CO2", 2019, 626159.9611543404],
                    ],
                    columns=["REGION", "EMISSION", "YEAR", "VALUE"],
                ).set_index(["REGION", "EMISSION", "YEAR"])
            },
        ),
    ]  # type: List

    @mark.parametrize(
        "results,input,expected",
        test_data,
        ids=["TotalDiscountedCost", "AnnualEmissions1"],
    )
    def test_convert_cbc_to_csv_long(self, results, input, expected):

        actual = convert_dataframe_to_csv(results, input)
        assert isinstance(actual, dict)
        for name, df in actual.items():
            pd.testing.assert_frame_equal(df, expected[name])

    def test_convert_cbc_to_csv_short(self):
        cbc_results = pd.DataFrame(
            data=[
                ["RateOfActivity", "SIMPLICITY,ID,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,IN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SD,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WD,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,ID,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,IN,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SD,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SN,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WD,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WN,DUMMY,1,2014", 1],
            ],
            columns=["Variable", "Index", "Value"],
        )
        input_data = {
            "EmissionActivityRatio": pd.DataFrame(
                data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
                columns=[
                    "REGION",
                    "TECHNOLOGY",
                    "EMISSION",
                    "MODE_OF_OPERATION",
                    "YEAR",
                    "VALUE",
                ],
            ).set_index(
                ["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"]
            ),
            "YearSplit": pd.DataFrame(
                data=[
                    ["ID", 2014, 0.1667],
                    ["IN", 2014, 0.0833],
                    ["SD", 2014, 0.1667],
                    ["SN", 2014, 0.0833],
                    ["WD", 2014, 0.3333],
                    ["WN", 2014, 0.1667],
                ],
                columns=["TIMESLICE", "YEAR", "VALUE"],
            ).set_index(["TIMESLICE", "YEAR"]),
        }

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "CO2", 2014, 1.0]],
            columns=["REGION", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "EMISSION", "YEAR"])

        actual = convert_dataframe_to_csv(cbc_results, input_data)
        assert isinstance(actual, dict)
        pd.testing.assert_frame_equal(actual["AnnualEmissions"], expected)


class TestCplexToCsv:

    test_data = [
        (
            "AnnualFixedOperatingCost	REGION	AOBACKSTOP	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0		",
            [],
        ),
        (
            "AnnualFixedOperatingCost	REGION	CDBACKSTOP	0.0	0.0	137958.8400384134	305945.38410619126	626159.9611543404	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0",
            [
                'AnnualFixedOperatingCost,"REGION,CDBACKSTOP,2017",137958.8400384134\n',
                'AnnualFixedOperatingCost,"REGION,CDBACKSTOP,2018",305945.3841061913\n',
                'AnnualFixedOperatingCost,"REGION,CDBACKSTOP,2019",626159.9611543404\n',
            ],
        ),
        (
            """RateOfActivity	REGION	S1D1	CGLFRCFURX	1	0.0	0.0	0.0	0.0	0.0	0.3284446367303371	0.3451714779880536	0.3366163200621617	0.3394945166233896	0.3137488154250392	0.28605725055560716	0.2572505015401749	0.06757558148965725	0.0558936625751148	0.04330608461292407	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0""",
            [
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2020",0.3284446367303371\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2021",0.3451714779880536\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2022",0.3366163200621617\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2023",0.3394945166233896\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2024",0.3137488154250392\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2025",0.28605725055560716\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2026",0.2572505015401749\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2027",0.06757558148965725\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2028",0.0558936625751148\n',
                'RateOfActivity,"REGION,S1D1,CGLFRCFURX,1,2029",0.04330608461292407\n',
            ],
        ),
    ]

    @mark.parametrize("cplex_input,expected", test_data)
    def test_convert_from_cplex_to_cbc(self, cplex_input, expected):

        actual = process_line(cplex_input, 2015, 2070, "csv")
        assert actual == expected


class TestCplexToCbc:

    test_data = [
        (
            "AnnualFixedOperatingCost	REGION	AOBACKSTOP	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0		",
            [],
        ),
        (
            "AnnualFixedOperatingCost	REGION	CDBACKSTOP	0.0	0.0	137958.8400384134	305945.38410619126	626159.9611543404	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0",
            [
                "0 AnnualFixedOperatingCost(REGION,CDBACKSTOP,2017) 137958.8400384134 0\n",
                "0 AnnualFixedOperatingCost(REGION,CDBACKSTOP,2018) 305945.3841061913 0\n",
                "0 AnnualFixedOperatingCost(REGION,CDBACKSTOP,2019) 626159.9611543404 0\n",
            ],
        ),
        (
            """RateOfActivity	REGION	S1D1	CGLFRCFURX	1	0.0	0.0	0.0	0.0	0.0	0.3284446367303371	0.3451714779880536	0.3366163200621617	0.3394945166233896	0.3137488154250392	0.28605725055560716	0.2572505015401749	0.06757558148965725	0.0558936625751148	0.04330608461292407	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0""",
            [
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2020) 0.3284446367303371 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2021) 0.3451714779880536 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2022) 0.3366163200621617 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2023) 0.3394945166233896 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2024) 0.3137488154250392 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2025) 0.28605725055560716 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2026) 0.2572505015401749 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2027) 0.06757558148965725 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2028) 0.0558936625751148 0\n",
                "0 RateOfActivity(REGION,S1D1,CGLFRCFURX,1,2029) 0.04330608461292407 0\n",
            ],
        ),
    ]

    @mark.parametrize("cplex_input,expected", test_data)
    def test_convert_from_cplex_to_cbc(self, cplex_input, expected):

        actual = process_line(cplex_input, 2015, 2070, "cbc")
        assert actual == expected



import pandas as pd
from amply import Amply

from otoole.preprocess.datafile_to_datapackage import (
    convert_amply_data_to_list,
    convert_amply_to_dataframe,
    load_parameter_definitions,
)


def test_amply():

    Amply(
        """set REGION;
          # set REGION := SIMPLICITY;
          set TECHNOLOGY;
          set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
          set MODE_OF_OPERATION;
          set MODE_OF_OPERATION := 1 2;
          set YEAR;
          set YEAR := 2014;
          end;"""
    )


def test_convert_amply_to_dataframe():

    config = {
        "VariableCost": {
            "type": "param",
            "indices": ["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"],
            "dtype": "float",
            "default": 0,
        },
        "REGION": {"type": "set", "dtype": "str"},
        "YEAR": {"dtype": "int", "type": "set"},
        "MODE_OF_OPERATION": {"dtype": "int", "type": "set"},
        "TECHNOLOGY": {"dtype": "str", "type": "set"},
    }

    amply = Amply(
        """set REGION;
                     set REGION := SIMPLICITY;
                     set TECHNOLOGY;
                     set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
                     set MODE_OF_OPERATION;
                     set MODE_OF_OPERATION := 1 2;
                     set YEAR;
                     set YEAR := 2014;"""
    )
    amply.load_string("param VariableCost {REGION,TECHNOLOGY,MODE_OF_OPERATION,YEAR};")
    #     amply.load_string("""param default 0 : VariableCost :=
    # SIMPLICITY ETHPLANT 1 2014 2.89
    # SIMPLICITY ETHPLANT 2 2014 999999.0
    # SIMPLICITY GAS_EXTRACTION 1 2014 7.5
    # SIMPLICITY GAS_EXTRACTION 2 2014 999999.0""")
    amply.load_string(
        """
param VariableCost default 0.0001 :=
[SIMPLICITY,ETHPLANT,*,*]:
2014 :=
1 2.89
2 999999.0
[SIMPLICITY,GAS_EXTRACTION,*,*]:
2014 :=
1 7.5
2 999999.0;"""
    )
    actual = convert_amply_to_dataframe(amply, config)
    expected = pd.DataFrame(
        data=[
            ["SIMPLICITY", "ETHPLANT", 1, 2014, 2.89],
            ["SIMPLICITY", "ETHPLANT", 2, 2014, 999999.0],
            ["SIMPLICITY", "GAS_EXTRACTION", 1, 2014, 7.5],
            ["SIMPLICITY", "GAS_EXTRACTION", 2, 2014, 999999.0],
        ],
        columns=["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR", "VALUE"],
    )

    pd.testing.assert_frame_equal(actual["VariableCost"], expected)


def test_convert_amply_data_to_list_of_lists():

    data = {
        "SIMPLICITY": {
            "ETHPLANT": {1.0: {2014.0: 2.89}, 2.0: {2014.0: 999999.0}},
            "GAS_EXTRACTION": {1.0: {2014.0: 7.5}, 2.0: {2014.0: 999999.0}},
        }
    }
    expected = [
        ["SIMPLICITY", "ETHPLANT", 1.0, 2014.0, 2.89],
        ["SIMPLICITY", "ETHPLANT", 2.0, 2014.0, 999999.0],
        ["SIMPLICITY", "GAS_EXTRACTION", 1.0, 2014.0, 7.5],
        ["SIMPLICITY", "GAS_EXTRACTION", 2.0, 2014.0, 999999.0],
    ]
    actual = convert_amply_data_to_list(data)
    assert actual == expected


def test_load_parameters():

    config = {"TestParameter": {"type": "param", "indices": ["index1", "index2"]}}

    actual = load_parameter_definitions(config)
    expected = "param TestParameter {index1,index2};\n"
    assert actual == expected


def test_load_sets():

    config = {"TestSet": {"type": "set"}}

    actual = load_parameter_definitions(config)
    expected = "set TestSet;\n"
    assert actual == expected



import logging
import sys
from abc import abstractmethod
from typing import TextIO

import pandas as pd
from datapackage import Package
from pandas_datapackage_reader import read_datapackage
from sqlalchemy import create_engine

from otoole import read_packaged_file
from otoole.preprocess.excel_to_osemosys import CSV_TO_EXCEL

logger = logging.getLogger(__name__)


class DataPackageTo(object):
    """Convert a data package to another format

    Arguments
    ---------
    datapackage: str
        The path to the databackage
    datafilepath: str
        The path to the destination file or folder
    sql: bool, default=False
        Flag to set whether the source datapackage is in sqlite format
    """

    def __init__(self, datapackage: str, datafilepath: str, sql: bool = False):

        self.datapackage = datapackage
        self.datafilepath = datafilepath
        self.sql = sql
        self.package = self._get_package()
        self.default_values = self._get_default_values()
        self.config = read_packaged_file("config.yaml", "otoole.preprocess")

    def _get_package(self):

        if self.sql:
            engine = create_engine("sqlite:///{}".format(self.datapackage))
            package = Package(storage="sql", engine=engine)
        else:
            package = read_datapackage(self.datapackage)  # typing: datapackage.Package

        return package

    def _get_default_values(self):
        default_resource = (
            self.package.pop("default_values").set_index("name").to_dict()
        )
        return default_resource["default_value"]

    def convert(self):
        """Perform the conversion from datapackage to destination format
        """

        handle = self._header()
        logger.debug(self.default_values)

        for name, df in self.package.items():
            logger.debug(name)

            if df.empty:
                columns = [x["name"] for x in df._metadata["schema"]["fields"]]
                df = pd.DataFrame(columns=columns)

            df = df.reset_index()
            if "index" in df.columns:
                df = df.drop(columns="index")

            logger.debug("Number of columns: %s, %s", len(df.columns), df.columns)
            if len(df.columns) > 1:
                default_value = self.default_values[name]
                self._write_parameter(df, name, handle, default=default_value)

            else:
                self._write_set(df, name, handle)

        self._footer(handle)

        handle.close()

    @abstractmethod
    def _header(self) -> TextIO:
        raise NotImplementedError()

    @abstractmethod
    def _write_parameter(
        self, df: pd.DataFrame, parameter_name: str, handle: TextIO, default: float
    ) -> pd.DataFrame:
        """Write parameter data"""
        raise NotImplementedError()

    @abstractmethod
    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        raise NotImplementedError()

    @abstractmethod
    def _footer(self, handle: TextIO):
        raise NotImplementedError()


class DataPackageToDatafile(DataPackageTo):
    """Write datapackage to a GMPL datafile
    """

    def _header(self):
        filepath = open(self.datafilepath, "w")
        msg = "# Model file written by *otoole*\n"
        filepath.write(msg)
        return filepath

    def _form_parameter(self, df: pd.DataFrame, default: float):

        df = df[df.VALUE != default]
        return df

    def _write_parameter(
        self, df: pd.DataFrame, parameter_name: str, handle: TextIO, default: float
    ):
        """Write parameter data to a GMPL datafile, omitting data with default value

        Arguments
        ---------
        filepath : StreamIO
        df : pandas.DataFrame
        parameter_name : str
        handle: TextIO
        default : int
        """
        df = self._form_parameter(df, default)
        handle.write("param default {} : {} :=\n".format(default, parameter_name))
        df.to_csv(path_or_buf=handle, sep=" ", header=False, index=False)
        handle.write(";\n")

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO):
        """Write set data to a GMPL datafile

        Arguments
        ---------
        df : pandas.DataFrame
        set_name : str
        handle: TextIO
        """
        handle.write("set {} :=\n".format(set_name))
        df.to_csv(path_or_buf=handle, sep=" ", header=False, index=False)
        handle.write(";\n")

    def _footer(self, handle: TextIO):
        handle.write("end;\n")
        handle.close()


class DataPackageToExcel(DataPackageTo):
    def _header(self):
        return pd.ExcelWriter(self.datafilepath, mode="w")

    def _form_parameter(
        self, df: pd.DataFrame, parameter_name: str, default: float
    ) -> pd.DataFrame:
        """Converts data into wide format

        Arguments
        ---------
        df: pd.DataFrame
        parameter_name: str
        default: float

        Returns
        -------
        pandas.DataFrame
        """

        if not df.empty:

            names = df.columns.to_list()
            if len(names) > 2:
                logger.debug(
                    "More than 2 columns for {}: {}".format(parameter_name, names)
                )
                rows = names[0:-2]
                columns = names[-2]
                values = names[-1]
                logger.debug("Rows: {}; columns: {}; values: {}", rows, columns, values)
                logger.debug("dtypes: {}".format(df.dtypes))
                pivot = pd.pivot_table(
                    df, index=rows, columns=columns, values=values, fill_value=default
                )
            elif len(names) == 2:
                logger.debug("Two columns for {}: {}".format(parameter_name, names))
                values = names[-1]
                rows = names[0:-2]
                logger.debug("Rows: {}; values: {}", rows, values)
                pivot = pd.pivot_table(
                    df, index=rows, values=values, fill_value=default
                )
            else:
                logger.debug("One column for {}: {}".format(parameter_name, names))
                pivot = df.copy()
                pivot = pivot.reset_index(drop=True)

        else:
            logger.debug("Dataframe {} is empty".format(parameter_name))
            pivot = df.copy()

        return pivot

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: pd.ExcelWriter,
        default: float,
    ):
        try:
            name = CSV_TO_EXCEL[parameter_name]
        except KeyError:
            name = parameter_name
        df = self._form_parameter(df, parameter_name, default)
        df.to_excel(handle, sheet_name=name, merge_cells=False)

    def _write_set(self, df: pd.DataFrame, set_name, handle: pd.ExcelWriter):
        df.to_excel(handle, sheet_name=set_name, merge_cells=False, index=False)

    def _footer(self, handle=pd.ExcelWriter):
        handle.close()


def convert_datapackage_to_datafile(path_to_datapackage, path_to_datafile):
    dp = DataPackageToDatafile(path_to_datapackage, path_to_datafile)
    dp.convert()


def convert_datapackage_to_excel(path_to_datapackage, path_to_excel):
    dp = DataPackageToExcel(path_to_datapackage, path_to_excel)
    dp.convert()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    path_to_datapackage = sys.argv[1]
    path_to_datafile = sys.argv[2]

    DataPackageToDatafile(path_to_datapackage, path_to_datafile)



#!/usr/bin/env python
# coding: utf-8
"""Extract data from spreadsheets and write to an OSeMOSYS datafile

Reads and writes the following OSeMOSYS parameters and sets to and from
files on disk.

Notes
=====

Sets
----
These are the standard sets::

    set YEAR;
    set TECHNOLOGY;
    set TIMESLICE;
    set FUEL;
    set EMISSION;
    set MODE_OF_OPERATION;
    set REGION;
    set SEASON;
    set DAYTYPE;
    set DAILYTIMEBRACKET;
    set FLEXIBLEDEMANDTYPE;
    set STORAGE;

All sets are written in a CSV file in one column of values with header of ``VALUE``.
For example, the CSV file for set YEAR::

    VALUE
    2015
    2016
    2017
    2018
    2019
    2020

Sets are written into the OSeMOSYS data file using the following syntax::

    set YEAR := 2014 2015 2016 2017 2018 2019 2020 ;


Parameters
----------
In general, parameters can be written into CSV files in narrow or wide
formats. In narrow format, the CSV file should look as follows::

    REGION,TIMESLICE,YEAR,VALUE
    SIMPLICITY,ID,2014,0.1667
    SIMPLICITY,IN,2014,0.0833
    SIMPLICITY,SD,2014,0.1667
    SIMPLICITY,SN,2014,0.0833
    SIMPLICITY,WD,2014,0.3333
    SIMPLICITY,WN,2014,0.1667


In wide format, the final index is transposed::

    REGION,TIMESLICE,2014,2015,...
    SIMPLICITY,ID,0.1667,0.1667
    SIMPLICITY,IN,0.0833,0.0833
    SIMPLICITY,SD,0.1667,0.1667
    SIMPLICITY,SN,0.0833,0.0833
    SIMPLICITY,WD,0.3333,0.3333
    SIMPLICITY,WN,0.1667,0.1667

Wide format is a bit nicer to use with spreadsheets,
as it allows you to more easily plot graphs of values, but narrow
format is a more flexible and easily manipulated data format.

Writing parameteters:

1-dimensional e.g. DiscountRate{r in REGION}

2-dimension e.g. YearSplit{l in TIMESLICE, y in YEAR}

n-dimensional e.g. DaysInDayType{ls in SEASON, ld in DAYTYPE, y in YEAR}


Global parameters::

    param YearSplit{l in TIMESLICE, y in YEAR};
    param DiscountRate{r in REGION};
    param DaySplit{lh in DAILYTIMEBRACKET, y in YEAR};
    param Conversionls{l in TIMESLICE, ls in SEASON};
    param Conversionld{l in TIMESLICE, ld in DAYTYPE};
    param Conversionlh{l in TIMESLICE, lh in DAILYTIMEBRACKET};
    param DaysInDayType{ls in SEASON, ld in DAYTYPE, y in YEAR};
    param TradeRoute{r in REGION, rr in REGION, f in FUEL, y in YEAR};
    param DepreciationMethod{r in REGION};


Demand parameters::

    param SpecifiedAnnualDemand{r in REGION, f in FUEL, y in YEAR};
    param SpecifiedDemandProfile{r in REGION, f in FUEL, l in TIMESLICE, y in YEAR};
    param AccumulatedAnnualDemand{r in REGION, f in FUEL, y in YEAR};


Performance parameters::

    param CapacityToActivityUnit{r in REGION, t in TECHNOLOGY};
    param TechWithCapacityNeededToMeetPeakTS{r in REGION, t in TECHNOLOGY};
    param CapacityFactor{r in REGION, t in TECHNOLOGY, l in TIMESLICE, y in YEAR};
    param AvailabilityFactor{r in REGION, t in TECHNOLOGY, y in YEAR};
    param OperationalLife{r in REGION, t in TECHNOLOGY};
    param ResidualCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};
    param InputActivityRatio{r in REGION, t in TECHNOLOGY, f in FUEL,
                             m in MODE_OF_OPERATION, y in YEAR};
    param OutputActivityRatio{r in REGION, t in TECHNOLOGY, f in FUEL,
                              m in MODE_OF_OPERATION, y in YEAR};


Technology Costs parameters::

    param CapitalCost{r in REGION, t in TECHNOLOGY, y in YEAR};
    param VariableCost{r in REGION, t in TECHNOLOGY, m in MODE_OF_OPERATION, y in YEAR};
    param FixedCost{r in REGION, t in TECHNOLOGY, y in YEAR};


Storage parameters::

    param TechnologyToStorage{r in REGION, t in TECHNOLOGY, s in STORAGE,
                              m in MODE_OF_OPERATION};
    param TechnologyFromStorage{r in REGION, t in TECHNOLOGY, s in STORAGE,
                              m in MODE_OF_OPERATION};
    param StorageLevelStart{r in REGION, s in STORAGE};
    param StorageMaxChargeRate{r in REGION, s in STORAGE};
    param StorageMaxDischargeRate{r in REGION, s in STORAGE};
    param MinStorageCharge{r in REGION, s in STORAGE, y in YEAR};
    param OperationalLifeStorage{r in REGION, s in STORAGE};
    param CapitalCostStorage{r in REGION, s in STORAGE, y in YEAR};
    param ResidualStorageCapacity{r in REGION, s in STORAGE, y in YEAR};


Capacity Constraints parameters::

    param CapacityOfOneTechnologyUnit{r in REGION, t in TECHNOLOGY, y in YEAR};
    param TotalAnnualMaxCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};
    param TotalAnnualMinCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};


Investment Constraints parameters::

    param TotalAnnualMaxCapacityInvestment{r in REGION, t in TECHNOLOGY, y in YEAR};
    param TotalAnnualMinCapacityInvestment{r in REGION, t in TECHNOLOGY, y in YEAR};


Activity Constraints parameters::

    param TotalTechnologyAnnualActivityUpperLimit{r in REGION, t in TECHNOLOGY,
                                                  y in YEAR};
    param TotalTechnologyAnnualActivityLowerLimit{r in REGION, t in TECHNOLOGY,
                                                  y in YEAR};
    param TotalTechnologyModelPeriodActivityUpperLimit{r in REGION, t in TECHNOLOGY};
    param TotalTechnologyModelPeriodActivityLowerLimit{r in REGION, t in TECHNOLOGY};


Reserve Margin parameters::

    param ReserveMarginTagTechnology{r in REGION, t in TECHNOLOGY, y in YEAR};
    param ReserveMarginTagFuel{r in REGION, f in FUEL, y in YEAR};
    param ReserveMargin{r in REGION, y in YEAR};


RE Generation Target parameters::

    param RETagTechnology{r in REGION, t in TECHNOLOGY, y in YEAR};
    param RETagFuel{r in REGION, f in FUEL, y in YEAR};
    param REMinProductionTarget{r in REGION, y in YEAR};


Emissions & Penalties parameters::

    param EmissionActivityRatio{r in REGION, t in TECHNOLOGY, e in EMISSION,
                                m in MODE_OF_OPERATION, y in YEAR};
    param EmissionsPenalty{r in REGION, e in EMISSION, y in YEAR};
    param AnnualExogenousEmission{r in REGION, e in EMISSION, y in YEAR};
    param AnnualEmissionLimit{r in REGION, e in EMISSION, y in YEAR};
    param ModelPeriodExogenousEmission{r in REGION, e in EMISSION};
    param ModelPeriodEmissionLimit{r in REGION, e in EMISSION};

"""
import csv
import logging
import os
from tempfile import TemporaryDirectory
from typing import Dict, List

import xlrd

from otoole import read_packaged_file
from otoole.preprocess.longify_data import main as longify

logger = logging.getLogger(__name__)


EXCEL_TO_CSV = {
    "TotalAnnualMaxCapacityInvestmen": "TotalAnnualMaxCapacityInvestment",
    "TotalAnnualMinCapacityInvestmen": "TotalAnnualMinCapacityInvestment",
    "TotalTechnologyAnnualActivityLo": "TotalTechnologyAnnualActivityLowerLimit",
    "TotalTechnologyAnnualActivityUp": "TotalTechnologyAnnualActivityUpperLimit",
    "TotalTechnologyModelPeriodActLo": "TotalTechnologyModelPeriodActivityLowerLimit",
    "TotalTechnologyModelPeriodActUp": "TotalTechnologyModelPeriodActivityUpperLimit",
}

CSV_TO_EXCEL = {v: k for k, v in EXCEL_TO_CSV.items()}


def read_config(path_to_user_config: str = None) -> Dict:
    """Reads the config file holding expected OSeMOSYS set and parameter dimensions

    Arguments
    ---------
    path_to_user_config : str, optional, default=None
        Optional path to a user defined configuration file

    Returns
    -------
    dict
    """
    if path_to_user_config:
        config = read_packaged_file(path_to_user_config, None)
    else:
        config = read_packaged_file("config.yaml", "otoole.preprocess")
    return config


def generate_csv_from_excel(input_workbook, output_folder):
    """Generate a folder of narrow, well formatted CSV files from a spreadsheet

    Arguments
    ---------
    input_workbook : str
        Path to spreadsheet containing OSeMOSYS data
    output_folder : str
        Path of the folder containing the csv files

    """
    work_book = xlrd.open_workbook(os.path.join(input_workbook))

    with TemporaryDirectory() as temp_folder:
        _csv_from_excel(work_book, temp_folder)
        longify(temp_folder, output_folder)

    work_book.release_resources()  # release the workbook-resources
    del work_book


def _csv_from_excel(workbook: xlrd.Book, output_folder: str):
    """Creates csv files from all sheets in a workbook

    Arguments
    ---------
    workbook : xlrd.Book
    output_folder : str
    """

    logger.debug("Generating CSVs from Excel %s", workbook)
    # Create all the csv files in a new folder
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)  # creates the csv folder

    # Iterate over each sheet in the workbook
    for sheet in workbook.sheets():  # typing: xlrd.book.Sheet

        name = sheet.name
        mod_name = _modify_names([name])
        logger.debug(mod_name)

        # Open the sheet name in the xlsx file and write it in csv format]
        filepath = os.path.join(output_folder, mod_name[0] + ".csv")
        with open(filepath, "w", newline="") as your_csv_file:
            wr = csv.writer(your_csv_file, quoting=csv.QUOTE_NONNUMERIC)

            for rownum in range(sheet.nrows):  # reads each row in the csv file
                row = _cast_to_integer(sheet.row_values(rownum))
                wr.writerow(row)


def _modify_names(sheet_names: List) -> List:
    """I change the name of the sheets in the xlsx file to match with the csv
    actual ones
    """
    modified_names = []
    for name in sheet_names:
        try:
            new_name = EXCEL_TO_CSV[name]
        except KeyError:
            new_name = name

        modified_names.append(new_name)

    return modified_names


def _cast_to_integer(row):
    """function to convert all float numbers to integers....need to check it!!
    """
    if all(isinstance(n, float) for n in row):
        converted_row = list(map(int, row))
    else:
        converted_row = row
    return converted_row
